import os
import glob as glob1
import shutil

__all__ = ['copy', 'exists', 'isdir', 'isfile', 'listdir', 'makedirs', 
           'path_join', 'remove', 'rename', 'stat', 'walk', 'glob'
           ]


def exists(path):
    """Determines whether a path exists or not.
    Args:
        path: string, a path, filepath or dirpath.
    Returns:
        True if the path exists, whether it's a file or a directory. 
        False if the path does not exist and there are no filesystem errors.
    """
    return os.path.exists(path)


def isdir(path):
    """Returns whether the path is a directory or not.
    
    Args:
        path: string, path to a potential directory.
    Returns:
        True, if the path is a directory; False otherwise.
    """
    return os.path.isdir(path)


def isfile(path):
    """Returns whether the path is a regular file or not.
    
    Args:
        path: string, path to a potential file.
    Returns:
        True, if the path is a regular file; False otherwise.
    """
    return os.path.isfile(path)


def listdir(path, full_path=False):
    """Returns a list of entries contained within a directory.
    
    Args:
        path: string, path to a directory.
    Returns:
        [filename1, filename2, ... filenameN] as strings.
    Raises:
        errors. NotFoundError if directory doesn't exist.
    """
    return [path_join(path, i) for i in os.listdir(path)] if full_path else os.listdir(path)


def copy(src, dst, overwrite=False):
    """Copies data from src to dst.
    
    1.copy file to file.  eg:'/data/A/a.txt' to '/data/B/b.txt' --> exist '/data/B/b.txt'
    2.copy file to directory.  eg:'/data/A/a.txt' to '/data/B'  --> exist '/data/B/a.txt'
    3.copy directory to directory. eg:'/data/A' to '/data/B'    --> exist '/data/B/A'

    Args:
        src: string, name of the file or directory whose contents need to be copied
        dst: string, name of the file or directory to which to copy to
        overwrite: boolean, Whether to overwrite the file if existing file.
    """
    assert exists(src), "src not exists."
    if isfile(src):
        if overwrite or not exists(dst):
            return shutil.copy(src, dst)
        else:
            if isdir(dst):
                if not exists(path_join(dst, path_join(src).split('/')[-1])):
                    return shutil.copy(src, dst)
    else: 
        makedirs(dst)
        path = path_join(dst, path_join(src).rstrip('/').split('/')[-1])
        if overwrite or not exists(path):
            remove(path)
            return shutil.copytree(src, path)


def makedirs(path):
    """Creates a directory and all parent/intermediate directories.
    
    Args:
        path: string, name of the directory to be created.
    """
    if not exists(path):
        os.makedirs(path)


def remove(path):
    """Deletes a directory or file.
    
    Args:
        path: string, a path, filepath or dirpath.
    Raises:
        errors. NotFoundError if directory or file doesn't exist.
    """
    if exists(path):
        if isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path)


def rename(src, dst, overwrite=False):
    """Rename or move a file / directory.
    
    Args:
        src: string, pathname for a file or dir.
        dst: string, pathname to which the file needs to be moved.
        overwrite: boolean, Whether to overwrite the file if existing file.
    """
    assert exists(src), "src not exists."
    if exists(dst):
        assert isdir(src)==isdir(dst), "src and dst should same type."
        assert isfile(src)==isfile(dst), "src and dst should same type."
        if overwrite:
            shutil.rmtree(dst)
            os.rename(src, dst)
    else:
        os.rename(src, dst)
    

def stat(path):
    """Returns file or directory statistics for a given path.
    
    Args:
        path: string, a path, filepath or dirpath.
    Returns:
        FileStatistics struct that contains information about the path.
    """
    return os.stat(path)


def walk(path, topdown=True, onerror=None):
    """Recursive directory tree generator for directories.
    
    Args:
        path: string, a Directory name
        topdown: bool, Traverse pre order if True, post order if False.
        onerror: optional handler for errors. Should be a function, 
                 it will be called with the error as argument. 
                 Rethrowing the error aborts the walk. 
                 Errors that happen while listing directories are ignored.
    
    Returns:
        Yields, Each yield is a 3-tuple: the pathname of a directory, 
        followed by lists of all its subdirectories and leaf files. 
        That is, each yield looks like: (dirname, [subdirname, subdirname, ...], [filename, filename, ...]). 
        Each item is a string.
    """
    return list(os.walk(path, topdown, onerror))


def path_join(path, *paths):
    """
    Join two or more pathname components, inserting '/' as needed.
    If any component is an absolute path, all previous path components 
    will be discarded.  An empty last part will result in a path that 
    ends with a separator.
    
    Args:
        path: file or directory
        *paths: file or directory
    Return:
        a new path
    """
    return eval(repr(os.path.join(path, *paths)).replace("\\", '/').replace("//", '/'))


def glob(pathname, iterable=False):
    """Returns a list of files that match the given pathname(s).
    
    The pathname are defined as strings. 
    Supported patterns are defined here. 
    Note that the pattern can be a Python iteratable of string patterns.
    
    Args:
        pathname: The glob pattern(s).
        iterable: Returns an iterable object.
    Return:
        A list of strings containing filenames that match the given pattern(s).
    """
    if iterable:
        return glob1.iglob(pathname)
    return glob1.glob(pathname)


from linora.image._image_noise import *

class ImageNoiseAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def mosaic(self, size=(80,80), block=0, axis=None, prob=0.3, p=None):
        """mosaic noise apply to image.

        Args:
            size: if int or float, xsize=ysize, how many mosaic blocks the image is divided into.
                  if 2-tuple, (xsize, ysize), 
                  if 4-tuple, xsize in (size[0], size[1]) and  ysize in (size[2], size[3]).
            block: mosaic area block
            axis: list or tuple, one or more mosaic center axis point.
            prob: probability of numbers of mosaci.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = mosaic(self.image, size, block, axis, prob, p=p)
        return self

    def noise_saltpepper(self, white_prob=0.025, black_prob=0.025, p=None):
        """Mask salt pepper noise apply to image.

        The salt-pepper noise is based on the signal-to-noise ratio of the image,
        randomly generating the pixel positions in some images all channel,
        and randomly assigning these pixels to 0 or 255.

        Args:
            white_prob: white pixel prob.
            black_prob: black pixel prob.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_saltpepper(self.image, white_prob, black_prob, p=p)
        return self

    def noise_gaussian(self, mean=0, std=30, wise='pixel', scale=1, prob=0.2, p=None):
        """gaussian distribution noise apply to image.

        Args:
            mean: if int or float, value is gaussian distribution mean.
                  if tuple or list, randomly picked in the interval `[mean[0], mean[1])`.
            std: if int or float, value is gaussian distribution std.
                 if tuple or list, randomly picked in the interval `[std[0], std[1])`.
            wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
            scale: if int or float, value multiply with noise.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_gaussian(self.image, mean, std, wise, scale, prob, p=p)
        return self

    def noise_laplace(self, mean=0, lam=30, wise='pixel', scale=1, prob=0.2, p=None):
        """laplace distribution noise apply to image.

        Args:
            mean: if int or float, value is laplace distribution mean.
                  if tuple or list, randomly picked in the interval `[mean[0], mean[1])`.
            lam: if int or float, value is laplace distribution lam.
                 if tuple or list, randomly picked in the interval `[lam[0], lam[1])`.
            wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
            scale: if int or float, value multiply with noise.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_laplace(self.image, mean, lam, wise, scale, prob, p=p)
        return self

    def noise_poisson(self, lam=30, wise='pixel', scale=1, prob=0.2, p=None):
        """poisson distribution noise apply to image.

        Args:
            lam: if int or float, value is poisson distribution lam.
                 if tuple or list, randomly picked in the interval `[lam[0], lam[1])`.
            wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
            scale: if int or float, value multiply with noise.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_poisson(self.image, lam, wise, scale, prob, p=p)
        return self


    def noise_uniform(self, lower=-50, upper=50, wise='pixel', scale=1, prob=0.2, p=None):
        """uniform distribution noise apply to image.

        Args:
            lower: if int or float, value is uniform distribution lower.
                   if tuple or list, randomly picked in the interval `[lower[0], lower[1])`.
            upper: if int or float, value is uniform distribution upper.
                   if tuple or list, randomly picked in the interval `[upper[0], upper[1])`.
            wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
            scale: if int or float, value multiply with noise.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_uniform(self.image, lower, upper, wise, scale, prob, p=p)
        return self

    def noise_speckle(self, wise='pixel', prob=0.2, p=None):
        """speckle noise apply to image.

        pixel = pixel+pixel*gaussian

        Args:
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_speckle(self.image, wise, prob, p=p)
        return self


    def noise_impulse(self, binomial=0.5, wise='pixel', prob=0.2, p=None):
        """impulse noise apply to image.

        Args:
            binomial: if int or float, value is binomial distribution prob.
                      if tuple or list, randomly picked in the interval `[binomial[0], binomial[1])`.
            wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying noise.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = noise_impulse(self.image, binomial, wise, prob, p=p)
        return self



import os
import bz2
import gzip
import lzma
import rarfile
import tarfile
import zipfile

import linora.gfile as gfile

__all__ = ['decompress', 'compress']


def decompress(file, folder=None):
    """Decompression file.
    
    Args:
        file: str, file should be file path;
        folder: str, decompression folder.
    Return:
        decompression folder.
    """
    mat = file.split('.')[-1]
    if folder is None:
        folder = file[:-(len(mat)+1)]
    if mat in ['gz']:
        with gzip.GzipFile(file) as g:
            with open(folder, "w+") as f:
                f.write(g.read())
    elif mat in ['tar', 'tgz']:
        with tarfile.open(file, 'r') as g:
            names = g.getnames()
            for name in names:
                g.extract(name, folder)
    elif mat in ['zip']:
        with zipfile.ZipFile(file, 'r') as g:
            names = g.namelist()
            for name in names:
                g.extract(name, folder)
    elif mat in ['rar']:
        with rarfile.RarFile(file, 'r') as g:
            names = g.namelist()
            for name in names:
                g.extract(name, folder)
    elif mat in ['bz2']:
        with bz2.BZ2File(file, 'r') as g:
            with open(folder, 'w') as f:
                f.write(g.read())
    elif mat in ['xz']:
        with lzma.open(file) as g:
            with open(folder, 'wb') as f:
                f.write(g.read())
    else:
        raise ValueError("`file` should be type of ['.gz', '.tar', '.tgz', '.zip', '.rar', '.bz2', '.xz'].")
    return folder


def compress(files, file):
    """Compress folder or file or list of files to file.
    
    Args:
        files: str or list
               if str, files should be file or folder path;
               if list, files should be file path list.
        file: str, compression files name.
    Return:
        compression files name.
    """
    if gfile.isdir(files):
        mat = file.split('.')[-1]
        if mat in ['zip']:
            with zipfile.ZipFile(file, 'w', zipfile.ZIP_DEFLATED) as z:
                for dirpath, dirnames, filenames in gfile.walk(files):
                    fpath = dirpath.replace(files, '')
                    fpath = fpath and fpath + os.sep or ''
                    for filename in filenames:
                        z.write(gfile.path_join(dirpath, filename), gfile.path_join(fpath, filename))
        elif mat in ['tar']:
            with tarfile.open(file, 'w') as tar:
                for dirpath, dirnames, filenames in gfile.walk(files):
                    fpath = dirpath.replace(files, '')
                    fpath = fpath and fpath + os.sep or ''
                    for filename in filenames:
                        tar.add(gfile.path_join(dirpath, filename), gfile.path_join(fpath, filename))
        else:
            raise ValueError("`file` should be type of ['.tar', '.zip'].")
    else:
        if isinstance(files, str):
            files = [files]
        assert isinstance(files, list), 'files should be file or list of files path.'
        for i in files:
            assert gfile.isfile(i), 'files should be file or list of files path.'
        mat = file.split('.')[-1]
        if mat in ['zip']:
            with zipfile.ZipFile(file, 'w', zipfile.ZIP_DEFLATED) as z:
                for i in files:
                    z.write(i)
        elif mat in ['tar']:
            with tarfile.TarFile(file, 'w') as t:
                for i in files:
                    t.add(i)
        elif mat in ['bz2']:
             with bz2.BZ2File(file, 'w') as b:
                for i in files:
                    with open(i, 'rb') as f:
                        b.write(f.read())
        elif mat in ['xz']:
            with open(file, 'wb') as fw:
                for i in files:
                    with open(i, 'rb') as fr:
                        fw.write(lzma.compress(fr.read()))
        else:
            raise ValueError("`file` should be type of ['.tar', '.zip', '.bz2', '.xz'].")
    return file


import json
import logging
import requests

from linora.utils._config import Config

__all__ = ['RemoteMonitor']


class RemoteMonitor():
    """Callback used to stream events to a server.

    Requires the `requests` library.
    Events are sent to `root + '/publish/epoch/end/'` by default. 
    Calls are HTTP POST, with a `data` argument which is a JSON-encoded dictionary of event data.
    If `send_as_json=True`, the content type of the request will be `"application/json"`.
    Otherwise the serialized JSON will be sent within a form.
    
    Args:
        root: String; root url of the target server.
        path: String; path relative to `root` to which the events will be sent.
        field: String; JSON field under which the data will be stored.
            The field is used only if the payload is sent within a form
            (i.e. send_as_json is set to False).
        headers: Dictionary; optional custom HTTP headers.
        send_as_json: Boolean; whether the request should be sent as `"application/json"`.
    """

    def __init__(self, root='http://localhost:9000', path='/publish/epoch/end/', 
                 field='data', headers=None, send_as_json=False):
        self._params = Config()
        self._params.root = root+path
        self._params.field = field
        self._params.headers = headers
        self._params.send_as_json = send_as_json
        self._params.name = 'RemoteMonitor'

    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        send = {'batch':batch}
        for k, v in log.items():
            send[k] = v.item() if isinstance(v, (np.ndarray, np.generic)) else v
        try:
            if self.send_as_json:
                requests.post(self._params.root, json=send, headers=self._params.headers)
            else:
                requests.post(self._params.root, {self._params.field:json.dumps(send)}, headers=self._params.headers)
        except requests.exceptions.RequestException:
            logging.warning(f'Warning: could not reach RemoteMonitor root server at {self.root}')



import colorsys

import numpy as np
from PIL import ImageEnhance, ImageOps, Image

__all__ = ['enhance_saturation', 'enhance_brightness', 'enhance_sharpness', 'enhance_contrast_sigmoid',
           'enhance_contrast_log', 'enhance_contrast_linear', 'enhance_contrast_gamma',
           'enhance_hue', 'color_invert', 'color_clip', 'equalize', 'rgb_hex',
           'hls_to_rgb', 'rgb_to_hls', 'hsv_to_rgb', 'rgb_to_hsv', 'rgb_to_yiq', 'yiq_to_rgb',
           'rgb_to_yuv', 'yuv_to_rgb', 'dropout', 'ydbdr_to_rgb', 'rgb_to_ydbdr', 'ycbcr_to_rgb', 
           'rgb_to_ycbcr', 'rgb_to_ypbpr', 'ypbpr_to_rgb', 'bgr_to_rgb', 'rgb_to_bgr',
           'rgb_to_grayscale', 'lab_to_rgb', 'rgb_to_lab', 'rgb_to_xyz', 'xyz_to_rgb'
          ]


def enhance_saturation(image, delta, p=1):
    """Adjust image color balance.

    This class can be used to adjust the colour balance of an image, 
    in a manner similar to the controls on a colour TV set. 
    An enhancement factor of 0.0 gives a black and white image. 
    A factor of 1.0 gives the original image.
    
    Args:
        image: a PIL instance.
        delta: A floating point value controlling the enhancement. 
               delta 1.0 always returns a copy of the original image, 
               lower factors mean less color, and higher values more. 
               There are no restrictions on this value.
               if list, tuple, randomly picked in the interval `[delta[0], delta[1])` , 
               value is float multiplier for adjusting color.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(delta, (list, tuple)):
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Color(image).enhance(delta)


def enhance_contrast_linear(image, delta, p=1):
    """Adjust image contrast.
  
    Contrast is adjusted independently for each channel of each image.
    pixel = (x - mean) * delta + mean
    
    Tips:
        1 means pixel value no change.
        0 means all pixel equal, pixel is image mean.
    Args:
        images: a PIL instance.
        delta: if int, float, a float multiplier for adjusting contrast.
               if list, tuple, randomly picked in the interval `[delta[0], delta[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(delta, (list, tuple)):
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Contrast(image).enhance(delta)


def enhance_contrast_gamma(image, gamma=1, gain=1.0, p=1):
    """Perform gamma correction on an image.
    
    For gamma greater than 1, the histogram will shift towards left and the output image will be darker than the input image. 
    For gamma less than 1, the histogram will shift towards right and the output image will be brighter than the input image.
    
    pixel = 255*gain *((pixel/255)**gamma)
    Values in the range gamma=(0.5, 2.0) seem to be sensible.
    
    Args:
        image: a PIL instance.
        gamma: float, Non negative real number, gamma larger than 1 make the shadows darker, 
               while gamma smaller than 1 make dark regions lighter.
               if list or tuple, randomly picked in the interval `[gamma[0], gamma[1])`.
        gain: float, The constant multiplier.
              if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(gamma, (list, tuple)):
        gamma = np.random.randint(gamma[0], gamma[1])
    if isinstance(gain, (list, tuple)):
        gain = np.random.uniform(gain[0], gain[1])
    return image.point(lambda x:int((255 + 1 - 1e-3) * gain * pow(x / 255.0, gamma)))


def enhance_contrast_sigmoid(image, cutoff=0.5, gain=10, p=1):
    """Perform sigmoid correction on an image.
    
    pixel = 255*1/(1+exp(gain*(cutoff-pixel/255)))
    Values in the range gain=(5, 20) and cutoff=(0.25, 0.75) seem to be sensible.
    
    Args:
        image: a PIL instance.
        cutoff: float, Cutoff that shifts the sigmoid function in horizontal direction. 
                Higher values mean that the switch from dark to light pixels happens later. 
                if list or tuple, randomly picked in the interval `[cutoff[0], cutoff[1])`.
        gain: float, The constant multiplier.
              if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(cutoff, (list, tuple)):
        cutoff = np.random.uniform(cutoff[0], cutoff[1])
    if isinstance(gain, (list, tuple)):
        gain = np.random.randint(gain[0], gain[1])
    return image.point(lambda x:int(255*1/(1+np.exp(gain*(cutoff-x / 255.0)))))


def enhance_contrast_log(image, gain=1, p=1):
    """Perform log correction on an image.
    
    pixel = 255*gain*log_2(1+pixel/255)
    Values in the range gain=[0.6, 1.4] seem to be sensible.
    
    Args:
        image: a PIL instance.
        gain: float, The constant multiplier.
              if list or tuple, randomly picked in the interval `[gain[0], gain[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(gain, (list, tuple)):
        gain = np.random.randint(gain[0], gain[1])
    return image.point(lambda x:int(255*gain*np.log2(1+x / 255.0)))


def enhance_brightness(image, delta, p=1):
    """Adjust image brightness.
    
    Tips:
        delta >1 to white, <1 to black.
        a suitable interval is [0.5, 1.5].
        0 means black.
        1 means pixel value no change.
    Args:
        image: a PIL instance.
        delta: if int, float, Amount to add to the pixel values.
               if list, tuple, randomly picked in the interval `[delta[0], delta[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(delta, (list, tuple)):
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Brightness(image).enhance(delta)


def enhance_sharpness(image, delta, p=1):
    """Adjust image sharpness.

    This class can be used to adjust the sharpness of an image. 
    An enhancement factor of 0.0 gives a blurred image, 
    a factor of 1.0 gives the original image, 
    and a factor of 2.0 gives a sharpened image.
    
    Args:
        image: a PIL instance.
        delta: A floating point value controlling the enhancement. 
               delta 1.0 always returns a copy of the original image, 
               lower factors mean less sharpness, and higher values more. 
               There are no restrictions on this value.
               if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` , value is float multiplier for adjusting color.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(delta, (list, tuple)):
        assert delta[0]<delta[1], 'delta should be delta[1] > delta[0]'
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Sharpness(image).enhance(delta)


def enhance_hue(image, delta, p=1):
    """Adjust image hue.
    
    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.
    
    `delta` is the amount of shift in H channel and must be in the interval `[-1, 1]`.

    Args:
        image: a PIL instance.
        delta: How much to shift the hue channel. Should be in [-1, 1]. 
               1 and -1 give complete reversal of hue channel in
               HSV space in positive and negative direction respectively.
               0 means no shift. Therefore, both -1 and 1 will give an image
               with complementary colors while 0 gives the original image.
               if list or tuple, randomly picked in the interval `[delta[0], delta[1])`.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(delta, (list, tuple)):
        delta = np.random.randint(delta[0], delta[1])
    assert -1. <= delta <= 1., "`delta` is not in [-1, 1]."
    input_mode = image.mode
    if input_mode in {"L", "1", "I", "F"}:
        return image
    h, s, v = image.convert("HSV").split()
    h = h.point(lambda x:x+delta * 127.5)
    return Image.merge("HSV", (h, s, v)).convert(input_mode)


def color_invert(image, lower=None, upper=None, wise='pixel', prob=1, p=1):
    """Invert colors of input PIL image.

    Args:
        image: a PIL instance.
        lower: int or list or tuple, [0, 255], All pixels below this greyscale level are inverted.
               if list or tuple, randomly picked in the interval `[lower[0], lower[1])`.
        upper: int or list or tuple, [0, 255], All pixels above this greyscale level are inverted.
               if list or tuple, randomly picked in the interval `[upper[0], upper[1])`.
        wise: 'pixel' or 'channel' or list of channel, method of applying operate.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(lower, (list, tuple)):
        lower = np.random.randint(lower[0], lower[1])
    if isinstance(upper, (list, tuple)):
        upper = np.random.randint(upper[0], upper[1])
    if lower is None and upper is None:
        lower = upper = 128
    elif lower is None:
        lower = 0
    elif upper is None:
        upper = 255
    if wise=='pixel':
        return image.point(lambda x: 255-x if (x<=lower or x>=upper) and np.random.uniform()<prob else x)
    elif wise=='channel':
        split = list(image.split())
        for i in range(len(split)):
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x: 255-x if x<=lower or x>=upper else x)
        return Image.merge(image.mode, split)
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        for i in wise:
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x: 255-x if x<=lower or x>=upper else x)
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` should be 'pixel' or 'channel' or list of channel.")


def color_clip(image, lower=None, upper=None, wise='pixel', prob=1, p=1):
    """Clip colors of input PIL image.
    
    Args:
        image: a PIL instance.
        lower: int or list or tuple, [0, 255], All pixels below this greyscale level are clipped.
               if list or tuple, randomly picked in the interval `[lower[0], lower[1])`.
        upper: int or list or tuple, [0, 255], All pixels above this greyscale level are clipped.
               if list or tuple, randomly picked in the interval `[upper[0], upper[1])`.
        wise: 'pixel' or 'channel' or list of channel, method of applying operate.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(lower, (list, tuple)):
        lower = np.random.randint(lower[0], lower[1])
    if isinstance(upper, (list, tuple)):
        upper = np.random.randint(upper[0], upper[1])
    if lower is None and upper is None:
        raise ValueError('`lower` and `upper` at least one value.')
    if wise=='pixel':
        if lower is not None:
            image = image.point(lambda x: lower if x<=lower and np.random.uniform()<prob else x)
        if upper is not None:
            image = image.point(lambda x: upper if x>=upper and np.random.uniform()<prob else x)
        return image
    elif wise=='channel':
        split = list(image.split())
        for i in range(len(split)):
            if np.random.uniform()<prob:
                if lower is not None:
                    split[i] = split[i].point(lambda x: lower if x<=lower else x)
                if upper is not None:
                    split[i] = split[i].point(lambda x: upper if x>=upper else x)
        return Image.merge(image.mode, split)
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        for i in wise:
            if np.random.uniform()<prob:
                if lower is not None:
                    split[i] = split[i].point(lambda x: lower if x<=lower else x)
                if upper is not None:
                    split[i] = split[i].point(lambda x: upper if x>=upper else x)
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` should be 'pixel' or 'channel' or list of channel.")


def equalize(image, p=1):
    """
    Equalize the image histogram. This function applies a non-linear
    mapping to the input image, in order to create a uniform
    distribution of grayscale values in the output image.

    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    return ImageOps.equalize(image)


def rgb_to_hsv(image):
    """Convert RGB image to HSV(Hue,Saturation,Value(Brightness)) image.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy HSV image with same type of image.
    """
    if image.max()>1:
        image = image/255.
    r, g, b = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_hsv = np.vectorize(colorsys.rgb_to_hsv)
    h, s, v = to_hsv(r, g, b)
    return np.stack((h, s, v), axis=2)


def hsv_to_rgb(image, normalize=False, dtype='float32'):
    """Convert HSV(Hue,Saturation,Value(Brightness)) image to RGB image.

    Args:
        image: NumPy HSV image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy HSV image with same shape of image.
    """
    norm = 255. if normalize else 1.
    h, s, v = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_rgb = np.vectorize(colorsys.hsv_to_rgb)
    r, g, b = to_rgb(h, s, v)
    return (np.stack((r, g, b), axis=2)*norm).astype(dtype)


def rgb_to_hls(image):
    """Convert RGB image to HLS(Hue,Lightness,Saturation) image.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy HLS image with same type of image.
    """
    if image.max()>1:
        image = image/255.
    r, g, b = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_hls = np.vectorize(colorsys.rgb_to_hls)
    h, l, s = to_hls(r, g, b)
    return np.stack((h, l, s), axis=2)


def hls_to_rgb(image, normalize=False, dtype='float32'):
    """Convert HLS(Hue,Lightness,Saturation) image to RGB image.

    Args:
        image (numpy.ndarray): NumPy HLS image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy HLS image with same shape of image.
    """
    norm = 255. if normalize else 1.
    h, l, s = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_rgb = np.vectorize(colorsys.hls_to_rgb)
    r, g, b = to_rgb(h, l, s)
    return (np.stack((r, g, b), axis=2)*norm).astype(dtype)


def rgb_to_yiq(image):
    """Convert RGB image to YIQ(Luminance,Chrominance) image.

    Args:
        image (numpy.ndarray): NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy YIQ image with same type of image.
    """
    if image.max()>1:
        image = image/255.
    r, g, b = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_yiq = np.vectorize(colorsys.rgb_to_yiq)
    y, i, q = to_yiq(r, g, b)
    return np.stack((y, i, q), axis=2)


def yiq_to_rgb(image, normalize=False, dtype='float32'):
    """Convert YIQ(Luminance,Chrominance) image to RGB image.

    Args:
        image: NumPy YIQ image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy YIQ image with same shape of image.
    """
    norm = 255. if normalize else 1.
    y, i, q = image[:, :, 0], image[:, :, 1], image[:, :, 2]
    to_rgb = np.vectorize(colorsys.yiq_to_rgb)
    r, g, b = to_rgb(y, i, q)
    return (np.stack((r, g, b), axis=2)*norm).astype(dtype)


def rgb_to_yuv(image):
    """Converts image from RGB to YUV.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array.
    """
    if image.max()>1:
        image = image/255.
    kernel = np.array([[0.299, -0.14714119, 0.61497538],
                       [0.587, -0.28886916, -0.51496512],
                       [0.114, 0.43601035, -0.10001026]])
    return np.dot(image, kernel)


def yuv_to_rgb(image, normalize=False, dtype='float32'):
    """Converts image from YUV to RGB.

    Args:
        image: NumPy YUV image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array.
    """
    norm = 255. if normalize else 1.
    kernel = np.array([[1, 1, 1], 
                       [0, -0.394642334, 2.03206185],
                       [1.13988303, -0.58062185, 0]])
    return (np.dot(image, kernel)*norm).astype(dtype)


def rgb_hex(color):
    """rgb to hexadecimal or hexadecimal to rgb.
    
    Args:
        color: str or list or tuple, if str, is hexadecimal; if list or tuple, is rgb value.
    return:
        if str, return rgb value; if list or tuple, return hexadecimal.
    """
    if isinstance(color, str):
        if len(color)==7 and color[0]=='#':
            return (int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16))
    if isinstance(color, (list, tuple)):
        return '#'+''.join([str(hex(i))[-2:].replace('x', '0').upper() for i in color])
    raise ValueError('`color` value error, should be str or list or tuple.')
    

def dropout(image, value=0, wise='pixel', prob=0.1, p=1):
    """Drop random channels from images.

    For image data, dropped channels will be filled with value.
    
    Args:
        image: a PIL instance.
        value: int or list, dropped channels will be filled with value.
        wise: 'pixel' or 'channel' or list of channel, method of applying operate.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    returns: 
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(value, (int, float)):
        value = [int(value)]
    if wise=='pixel':
        return image.point(lambda x:np.random.choice(value) if np.random.uniform()<prob else x)
    elif wise=='channel':
        split = list(image.split())
        index = np.random.randint(0, len(split))
        for i in range(len(split)):
            if np.random.uniform()<prob and i!=index:
                split[i] = split[i].point(lambda x:np.random.choice(value))
        return Image.merge(image.mode, split)
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        index = np.random.randint(0, len(split))
        if index in wise:
            if len(split)==len(wise):
                index = 0
            else:
                index = -1
        for i in wise:
            if np.random.uniform()<prob and i!=index:
                split[i] = split[i].point(lambda x:np.random.choice(value))
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` value error.")


def rgb_to_bgr(image):
    """Convert RGB image to BGR image.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy BGR image with same type of image.
    """
    return np.stack((image[:,:,2], image[:,:,1], image[:,:,0]), axis=2)


def bgr_to_rgb(image):
    """Convert BGR image to RGB image.

    Args:
        image: NumPy BGR image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy BGR image with same shape of image.
    """
    return rgb_to_bgr(image)


def rgb_to_ypbpr(image):
    """Convert BGR image to YPbPr image.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy YPbPr image with same shape of image.
    """
    kernel = np.array([[0.299, 0.587, 0.114],
                       [-0.168736, -0.331264, 0.5],
                       [0.5, -0.418688, -0.081312]])
    return np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))


def ypbpr_to_rgb(image, normalize=False, dtype='float32'):
    """Convert a YPbPr image to RGB image.
    
    Args:
        image: NumPy YPbPr image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy RGB image with same shape of image.
    """
    norm = 255. if normalize else 1.
    kernel = np.array([[1.00000000e00, -1.21889419e-06, 1.40199959e00],
                       [1.00000000e00, -3.44135678e-01, -7.14136156e-01],
                       [1.00000000e00, 1.77200007e00, 4.06298063e-07]])
    return (np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))*norm).astype(dtype)


def rgb_to_ycbcr(image):
    """Convert a RGB image to YCbCr image.
    
    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy YCbCr image with same shape of image.
    """
    if image.max()>1:
        image = image/255.
    image = rgb_to_ypbpr(image)
    return image*np.array([219, 224, 224], dtype=image.dtype)+np.array([16, 128, 128], dtype=image.dtype)


def ycbcr_to_rgb(image, normalize=False, dtype='float32'):
    """Convert a YCbCr image to RGB image.
    
    Args:
        image: NumPy YCbCr image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy RGB image with same shape of image.
    """
    norm = 255. if normalize else 1.
    image = image-np.array([16, 128, 128], dtype=image.dtype)
    image = image/np.array([219, 224, 224], dtype=image.dtype)
    return (ypbpr_to_rgb(image)*norm).astype(dtype)


def rgb_to_ydbdr(image):
    """Convert BGR image to YDbDr image.

    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy YDbDr image with same shape of image.
    """
    kernel = np.array([[0.299, 0.587, 0.114], [-0.45, -0.883, 1.333], [-1.333, 1.116, 0.217]])
    return np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))


def ydbdr_to_rgb(image, normalize=False, dtype='float32'):
    """Convert a YDbDr image to RGB image.
    
    Args:
        image: NumPy YDbDr image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy RGB image with same shape of image.
    """
    kernel = np.array([[1.00000000e00, 9.23037161e-05, -5.25912631e-01],
                       [1.00000000e00, -1.29132899e-01, 2.67899328e-01],
                       [1.00000000e00, 6.64679060e-01, -7.92025435e-05]])
    return np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))


def rgb_to_xyz(image):
    """Convert a RGB image to CIE XYZ image.
    
    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy CIE XYZ image with same shape of image.
    """
    kernel = np.array([[0.412453, 0.357580, 0.180423],
                       [0.212671, 0.715160, 0.072169],
                       [0.019334, 0.119193, 0.950227]])
    image = np.where(np.greater(image, 0.04045), np.power((image+0.055)/1.055, 2.4), image/12.92)
    return np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))


def xyz_to_rgb(image, normalize=False, dtype='float32'):
    """Convert a CIE XYZ image to RGB image.
    
    Args:
        image: NumPy CIE XYZ image array of shape (H, W, C) to be converted.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy RGB image with same shape of image.
    """
    norm = 255. if normalize else 1.
    kernel = np.array([[3.24048134, -1.53715152, -0.49853633],
                       [-0.96925495, 1.87599, 0.04155593],
                       [0.05564664, -0.20404134, 1.05731107]])
    image = np.tensordot(image, kernel.transpose(), axes=((-1,), (0,)))
    image = np.where(np.greater(image, 0.0031308), np.power(image, 1.0/2.4)*1.055-0.055, image*12.92)
    return (np.clip(image, 0, 1)*norm).astype(dtype)


def rgb_to_lab(image, illuminant="D65", observer="2"):
    """Convert a RGB image to CIE LAB image.
    
    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
        illuminant : {"A", "D50", "D55", "D65", "D75", "E"}, the name of the illuminant.
        observer : {"2", "10"}, the aperture angle of the observer.
    Returns:
        a numpy array, NumPy CIE LAB image with same shape of image.
    """
    illuminants = {
        "A": {
            "2": (1.098466069456375, 1, 0.3558228003436005),
            "10": (1.111420406956693, 1, 0.3519978321919493),
        },
        "D50": {
            "2": (0.9642119944211994, 1, 0.8251882845188288),
            "10": (0.9672062750333777, 1, 0.8142801513128616),
        },
        "D55": {
            "2": (0.956797052643698, 1, 0.9214805860173273),
            "10": (0.9579665682254781, 1, 0.9092525159847462),
        },
        "D65": {
            "2": (0.95047, 1.0, 1.08883),
            "10": (0.94809667673716, 1, 1.0730513595166162),
        },
        "D75": {
            "2": (0.9497220898840717, 1, 1.226393520724154),
            "10": (0.9441713925645873, 1, 1.2064272211720228),
        },
        "E": {"2": (1.0, 1.0, 1.0), "10": (1.0, 1.0, 1.0)},
    }
    coords = np.array(illuminants[illuminant.upper()][observer])
    xyz = rgb_to_xyz(image)
    xyz = xyz / coords
    xyz = np.where(np.greater(xyz, 0.008856), np.power(xyz, 1.0 / 3.0), xyz * 7.787 + 16.0 / 116.0)
    x, y, z = xyz[:,:,0], xyz[:,:,1], xyz[:,:,2]
    l = (y * 116.0) - 16.0
    a = (x - y) * 500.0
    b = (y - z) * 200.0
    return np.stack([l, a, b], axis=2)


def lab_to_rgb(image, illuminant="D65", observer="2", normalize=False, dtype='float32'):
    """Convert a CIE LAB image to RGB image.
    
    Args:
        image: NumPy CIE LAB image array of shape (H, W, C) to be converted.
        illuminant : {"A", "D50", "D55", "D65", "D75", "E"}, the name of the illuminant.
        observer : {"2", "10"}, the aperture angle of the observer.
        normalize: if True, rgb numpy array is [0,255], if False, rgb numpy array is [0,1]
        dtype: rgb numpy array dtype.
    Returns:
        a numpy array, NumPy RGB image with same shape of image.
    """
    norm = 255. if normalize else 1.
    l, a, b = image[:,:,0], image[:,:,1], image[:,:,2]

    y = (l + 16.0) / 116.0
    x = (a / 500.0) + y
    z = y - (b / 200.0)

    z = np.maximum(z, 0)

    xyz = np.stack([x, y, z], axis=2)
    xyz = np.where(np.greater(xyz, 0.2068966), np.power(xyz, 3.0), (xyz - 16.0 / 116.0) / 7.787)

    illuminants = {
        "A": {
            "2": (1.098466069456375, 1, 0.3558228003436005),
            "10": (1.111420406956693, 1, 0.3519978321919493),
        },
        "D50": {
            "2": (0.9642119944211994, 1, 0.8251882845188288),
            "10": (0.9672062750333777, 1, 0.8142801513128616),
        },
        "D55": {
            "2": (0.956797052643698, 1, 0.9214805860173273),
            "10": (0.9579665682254781, 1, 0.9092525159847462),
        },
        "D65": {
            "2": (0.95047, 1.0, 1.08883),
            "10": (0.94809667673716, 1, 1.0730513595166162),
        },
        "D75": {
            "2": (0.9497220898840717, 1, 1.226393520724154),
            "10": (0.9441713925645873, 1, 1.2064272211720228),
        },
        "E": {"2": (1.0, 1.0, 1.0), "10": (1.0, 1.0, 1.0)},
    }
    coords = np.array(illuminants[illuminant.upper()][observer])
    return (xyz_to_rgb(xyz * coords)*norm).astype(dtype)


def rgb_to_grayscale(image):
    """Convert a RGB image to grayscale image.
    
    Args:
        image: NumPy RGB image array of shape (H, W, C) to be converted.
    Returns:
        a numpy array, NumPy grayscale image with same shape of image.
    """
    return np.expand_dims(np.tensordot(image, np.array([0.2125, 0.7154, 0.0721]), (-1, -1)), -1)


# import random
import itertools
# from functools import reduce

import numpy as np
import pandas as pd

from linora.sample._sampling import sampling_stratify

__all__ = ['kfold', 'train_test_split']


def kfold(df, stratify=None, n_splits=3, shuffle=True, seed=None):
    """K-Folds cross-validator
    
    Provides train/test indices to split data in train/test sets. 
    Split dataset into k consecutive folds (without shuffling by default).
    Each fold is then used once as a validation while the k - 1 remaining folds form the training set.
    
    Args:
        df: pd.DataFrame with shape (n_samples, n_features) or pd.Series or np.array or list,
            Training data, where n_samples is the number of samples and n_features is the number of features.
        stratify : pd.Series, shape (n_samples,)
            The target variable for supervised learning problems.
        n_splits : int, default=3
            Number of folds. Must be at least 2.
        shuffle: boolean, default=True, optional
            Whether to shuffle the data before splitting into batches.
        seed: int, random seed. Used when `shuffle` == True.
    Returns:
        list, length=n_splits, List each list containing train-test split of inputs.
    """
    fold = []
    for i in range(n_splits-1):
        index = list(itertools.chain.from_iterable(fold))
        fold.append(sampling_stratify(df.drop(index), stratify=stratify, frac=1/(n_splits-i), seed=seed))
        
    fold_end_train = list(itertools.chain.from_iterable(fold))
    fold_end_test = df.drop(fold_end_train).index.tolist()
    
    if shuffle:
        fold = [[pd.Series(df.drop(i).index.tolist()).sample(frac=1, random_state=seed).tolist(), i] for i in fold]
        fold.append([fold_end_train, pd.Series(fold_end_test).sample(frac=1, random_state=seed).tolist()])
    else:
        fold = [df.drop(i).index.tolist() for i in fold]
        fold = [[i, df.drop(i).index.tolist()] for i in fold]
        fold.append([df.drop(fold_end_test).index.tolist(), fold_end_test])
    return fold
#     t = df.sample(frac=1, random_state=seed).index if shuffle else df.index
#     if stratify is None:
#         m = int(np.floor(len(t)/n_splits))
#         fold = [t[i*m:(i+1)*m].tolist() for i in range(n_splits-1)]+[t[(n_splits-1)*m:].tolist()]
#     else:
#         t = stratify[t]
#         fold = []
#         for label in t.unique():
#             a = t[t==label].index
#             m = int(np.floor(len(a)/n_splits))
#             fold.append([a[i*m:(i+1)*m].tolist() for i in range(n_splits-1)]+[a[(n_splits-1)*m:].tolist()])
#         t = [[] for i in range(n_splits)]
#         for i in fold:
#             for j in range(n_splits):
#                 t[j] = t[j]+i[j]
#         fold = t.copy()
#     t = [[] for i in range(n_splits)]
#     for i in range(n_splits):
#         j = fold.copy()
#         j.pop(i)
#         j = reduce(lambda x,y:x+y, j)
#         t[i] = t[i]+[j, fold[i]]
#     return t


def train_test_split(df, stratify=None, test_size=0.2, shuffle=True, seed=None):
    """Split DataFrame or matrices into random train and test subsets
    
    Args:
        df: pd.DataFrame, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples and n_features is the number of features.
        stratify : pd.Series, shape (n_samples,)
            The target variable for supervised learning problems.
        test_size: float, optional (default=0.2)
            Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split.
        shuffle  : boolean, default=True, optional
            Whether to shuffle the data before splitting into batches.
        seed: int, random seed. Used when `shuffle` == True.
    
    Returns:
        list, length=2, List containing train-test split of inputs.
    """
    test = sampling_stratify(df, stratify=stratify, frac=test_size, seed=seed)
    train = df.drop(test).index.tolist()
    if shuffle:
        train = pd.Series(train).sample(frac=1, random_state=seed).tolist()
    else:
        test = df.drop(train).index.tolist()
    return [train, test]
    
#     t = df.sample(frac=1, random_state=seed).index if shuffle else df.index
#     if stratify is None:
#         t = [t[0:round(len(t)*(1-test_size))].tolist(), t[round(len(t)*(1-test_size)):].tolist()]
#     else:
#         t = stratify[t]
#         fold = []
#         for label in t.unique():
#             a = t[t==label].index
#             fold.append([a[0:round(len(a)*(1-test_size))].tolist(), a[round(len(a)*(1-test_size)):].tolist()])
#         t = [[], []]
#         for i in fold:
#             for j in range(2):
#                 t[j] = t[j]+i[j]
#     if shuffle:
#         random.shuffle(t[0], lambda :0.5)
#         random.shuffle(t[1], lambda :0.5)
#     return t



import numpy as np

class Hlines():
    def add_hlines(self, name, ydata, xmin, xmax, **kwargs):
        """Plot horizontal lines at each *y* from *xmin* to *xmax*.
        
        Args:
            name: data name.
            ydata: y-axis data, float or array-like.
            xmin, xmax : float or array-like
                Respective beginning and end of each line. If scalars are
                provided, all lines will have same length.
            colors : list of colors, default: :rc:`lines.color`
            linestyles : {'solid', 'dashed', 'dashdot', 'dotted'}, optional
            label : str, default: ''
        """
        if isinstance(ydata, (int, float)):
            ydata = [ydata]
        if 'colors' not in kwargs and 'colors' not in kwargs and 'linecolor' not in kwargs:
            kwargs['colors'] = [self._params.color.pop(0)[1]]*len(ydata)
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xmin'] = xmin
        self._params.ydata[name]['xmax'] = xmax
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'hlines'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_hlines
        return self
    
    def _execute_plot_hlines(self, fig, ax, i, j):
        ax_plot = ax.hlines(j['ydata'], j['xmin'], j['xmax'], **j['kwargs'])


import numpy as np


class Scatter():
    def add_scatter(self, name, xdata, ydata, **kwargs):
        """A scatter plot of *y* vs. *x* with varying marker size and/or color.
        
        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
            marksize: float or array-like, shape (n, ), The marker size in points**2.
            markcolor: array-like or list of colors or color, optional
                The marker colors. Possible values:

                - A scalar or sequence of n numbers to be mapped to colors using *cmap* and *norm*.
                - A 2D array in which the rows are RGB or RGBA.
                - A sequence of colors of length n.
                - A single color format string.

                Note that *c* should not be a single numeric RGB or RGBA sequence
                because that is indistinguishable from an array of values to be
                colormapped. If you want to specify the same RGB or RGBA value for
                all points, use a 2D array with a single row.  Otherwise, value-
                matching will have precedence in case of a size matching with *x* and *y*.

                If you wish to specify a single color for all points prefer the *color* keyword argument.

                Defaults to `None`. In that case the marker color is determined
                by the value of *color*, *facecolor* or *facecolors*. In case
                those are not specified or `None`, the marker color is determined
                by the next color of the ``Axes``' current "shape and fill" color cycle.
            mark: marker style string, 
                {'.': 'point', ',': 'pixel', 'o': 'circle', 'v': 'triangle_down', '^': 'triangle_up', 
                '<': 'triangle_left', '>': 'triangle_right', '1': 'tri_down', '2': 'tri_up', '3': 'tri_left', 
                '4': 'tri_right', '8': 'octagon', 's': 'square', 'p': 'pentagon', '*': 'star', 'h': 'hexagon1', 
                'H': 'hexagon2', '+': 'plus', 'x': 'x', 'D': 'diamond', 'd': 'thin_diamond', '|': 'vline', 
                '_': 'hline', 'P': 'plus_filled', 'X': 'x_filled', 0: 'tickleft', 1: 'tickright', 2: 'tickup', 
                3: 'tickdown', 4: 'caretleft', 5: 'caretright', 6: 'caretup', 7: 'caretdown', 8: 'caretleftbase', 
                9: 'caretrightbase', 10: 'caretupbase', 11: 'caretdownbase', 
                'None': 'nothing', None: 'nothing', ' ': 'nothing', '': 'nothing'}
            cmap: {'viridis', 'jet'}. *cmap* is only used if *color* is an array of floats.
            norm: If *color* is an array of floats, *norm* is used to scale the color data, 
                  *color*, in the range 0 to 1, in order to map into the colormap *cmap*.
            vmin, vmax : float, default: None
                *vmin* and *vmax* are used in conjunction with the default norm to
                map the color array *color* to the colormap *cmap*. If None, the
                respective min and max of the color array is used.
                It is deprecated to use *vmin*/*vmax* when *norm* is given.
            alpha : float, default: None, The alpha blending value, between 0 (transparent) and 1 (opaque).
            markedgewidth : float or array-like, The linewidth of the marker edges. 
                Note: The default *edgecolors* is 'face'. You may want to change this as well.
            markedgecolor : {'face', 'none', *None*} or color or sequence of color
                The edge color of the marker. Possible values:

                - 'face': The edge color will always be the same as the face color.
                - 'none': No patch boundary will be drawn.
                - A color or sequence of colors.

                For non-filled markers, *edgecolors* is ignored. Instead, the color
                is determined like with 'face', i.e. from *c*, *colors*, or *facecolors*.

            plotnonfinite : bool, default: False
                Whether to plot points with nonfinite *color* (i.e. ``inf``, ``-inf``
                or ``nan``). If ``True`` the points are drawn with the *bad*
                colormap color (see `.Colormap.set_bad`).
        """
#         if 'pointcolor' in kwargs or not self._params.set_label:
#             self._params.set_label = False
#         self._params.colorbar.add('viridis' if 'cmap' not in kwargs else kwargs['cmap'])
        
        if 'mark' in kwargs:
            kwargs['marker'] = kwargs.pop('mark')
        if 'marksize' in kwargs:
            kwargs['s'] = kwargs.pop('marksize')
        if 'markedgewidth' in kwargs:
            kwargs['linewidths'] = kwargs.pop('markedgewidth')
        if 'markeredgecolor' in kwargs:
            kwargs['edgecolors'] = kwargs.pop('markeredgecolor')
        if 'markcolor' not in kwargs:
            kwargs['c'] = [self._params.color.pop(0)[1]]*len(ydata)
        elif isinstance(kwargs['markcolor'], dict):
            kwargs['c'] = kwargs.pop('markcolor')['mode']
        else:
            kwargs['c'] = kwargs.pop('markcolor')
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'scatter'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_scatter
        return self
    
    def _execute_plot_scatter(self, fig, ax, i, j):
        ax_plot = ax.scatter(j['xdata'], j['ydata'], **j['kwargs'])
        ax_plot.set_label(i)
#         if not self._params.set_label:
#             if len(self._params.colorbar)>0:
#                 fig.colorbar(ax_plot)
#                 self._params.colorbar.remove(list(self._params.colorbar)[0])